/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All Rights Reserved.
 */

// Santiago Pericas-Geertsen <santiago.pericasgeertsen@oracle.com>

var System = Java.type('java.lang.System');
var when = require("when");
var avatar = require('org/glassfish/avatar');
var state = avatar.application.state;
var common = require('org/glassfish/avatar/common');
var Promise = common.Promise;
var avatarModel = require("org/glassfish/avatar/model");

module.exports = (function() {

    System.out.println("RestJs: Initializing 'db' backend using model-store API.");

    var store = avatarModel.newStore('derby', {
        database: 'test',
        username: 'APP',
        password: 'localhost',
        createDb: true,
        dropTables: true
    });

    var Family = avatarModel.newModel('family', {
        "name": {
            type: "string",
            primary: true
        },
        "description": "string"
    });

    var Product = avatarModel.newModel('product', {
        "name": {
            type: "string",
            primary: true
        },
        "madeBy": "string",
        "price": "number",
        "quantity": "integer"
    });

    Family.hasMany(Product, {
        as: 'products',
        foreign: 'family'
    });

    store.bind(Family, Product);

    System.out.println("Before connecting");

    store.connect().then(function() {
        var families = [];
        for (var i = 0; i < 10; i++) {
            families.push({
                name: i,
                description: "Widget Family " + i
            });
        }
        when.all(families.map(function(o) {
            return Family.create(o).then(function(f) {
                var products = [];
                for (var i = 0; i < 10; i++) {
                    products.push({
                        name: f.name * 10 + i,
                        madeBy: 'acme',
                        price: i + 1,
                        quantity: i + 10
                    });
                }
                when.all(products.map(function(o) {
                    return f.products.create(o).then(function(p) {
                        System.out.println("    Creating product " + (p.name) + " in family " + f.name);
                        return p;
                    });
                }));
                return f;
            });
        }));
    });

    var res = {
        Family: {},
        Product: {}
    };

    res.Family.find = function(fname) {
        var prefix = "family.";
        var f = state.get(prefix + fname);
        if (f) {
            // If in local cache, just return it
            System.out.println("Resolving family " + fname + " from cache");
            return when.resolve(f);
        } else {
            // Otherwise, fetch from db and update cache
            System.out.println("Fetching family " + fname + " from database");
            var deferred = Promise.defer();
            Family.get(fname).then(function(f) {
                state.put(prefix + fname, f);
                System.out.println("Updating family " + fname + " in cache");
                deferred.resolve(f);
            }).otherwise(function() {
                deferred.reject();
            });
            return deferred.promise;
        }
    };

    res.Family.findAll = function() {
        var prefix = "families";
        var fs = state.get(prefix);
        if (fs) {
            // If in local cache, just return it
            System.out.println("Resolving families from cache");
            return when.resolve(fs);
        } else {
            // Otherwise, fetch from db and update cache
            System.out.println("Fetching families from database");
            var deferred = Promise.defer();
            Family.getAll().then(function(fs) {
                state.put(prefix, fs);
                System.out.println("Updating families in cache");
                deferred.resolve(fs);
            }).otherwise(function() {
                deferred.reject();
            });
            return deferred.promise;
        }
    };

    res.Family.findProducts = function(fname) {
        var prefix = "family.products.";
        var ps = state.get(prefix + fname);
        if (ps) {
            // If in local cache, just return it
            System.out.println("Resolving family " + fname + " products from cache");
            return when.resolve(ps);
        } else {
            // Otherwise, fetch from db and update cache
            System.out.println("Fetching family " + fname + " products from database");
            var deferred = Promise.defer();
            Family.get(fname).then(function(f) {
                f.products.getAll().then(function(ps) {
                    state.put(prefix + fname, ps);
                    System.out.println("Updating family " + fname + " products in cache");
                    deferred.resolve(ps);
                });
            }).otherwise(function() {
                deferred.reject();
            });
            return deferred.promise;
        }
    };
    
    res.Product.find = function(pname) {
        var prefix = "product.";
        var p = state.get(prefix + pname);
        if (p) {
            // If in local cache, just return it
            System.out.println("Resolving product " + pname + " from cache");
            return when.resolve(p);
        } else {
            // Otherwise, fetch from db and update cache
            System.out.println("Fetching product " + pname + " from database");
            var deferred = Promise.defer();
            Product.get(pname).then(function(p) {
                state.put(prefix + pname, p);
                System.out.println("Updating product " + pname + " in cache");
                deferred.resolve(p);
            }).otherwise(function() {
                deferred.reject();
            });
            return deferred.promise;
        }
    };
    
    res.Family.create = function(family) {
        var defered = Promise.defer();
        this.find(family.name).then(function(f) {
            defered.reject("Family with name " + f.name + " already exists.");
        }).otherwise(function() {
            // new family.. create one
            // make sure family is a number...
            if(isNaN(parseInt(family.name))) {
                defered.reject("Family name should be an integer number.");
            } else {
                Family.create(family).then(function(f) {
                    var products = [];
                    for (var i = 0; i < 10; i++) {
                        products.push({
                            name: f.name * 10 + i,
                            madeBy: 'Acme',
                            price: i + 1,
                            quantity: i + 10
                        });
                    }
                    when.all(products.map(function(o) {
                        return f.products.create(o).then(function(p) {
                            System.out.println("    Created product " + (p.name) + " in family " + f.name);
                            return p;
                        });
                    })).then(function() {
                        //remove families from cache so that it is populated anew
                        state.remove("families");
                        defered.resolve();
                    }).otherwise(function(error) {
                        if(error) {
                            System.out.println("Error in creating products.. " + error);
                        }
                        defered.reject("Error in creating products for family " + f.name);
                    });
                }).otherwise(function() {
                    defered.reject("Error in creating Family " + family.name);
                });
            }
        });
        return defered.promise;
    };
    return res;
})();