/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of the GNU General
 * Public License Version 2 only ("GPL"). You may not use this file except
 * in compliance with the License.  You can obtain a copy of the License at
 * https://avatar.java.net/license.html or legal/LICENSE.txt.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 */

'use strict';

var common = require('org/glassfish/avatar/common');
var Promise = common.Promise;
var wrapCallback = common.wrapCallback;
var nodeCall = common.nodeCall;
var tryAndCatch = common.tryAndCatch;
var log = common.log;

var registry = {
    restServices: {},
    pushServices: {},
    socketServices: {},
    hasUrl: function (url) {
        return this.restServices[url] || this.pushServices[url] || this.socketServices[url];
    }
};

var serviceMatcher = function (path, services, exchanger) {
    for (var pathTemplate in registry[services]) {
        var params = matchUrlTemplate(pathTemplate, path, registry[services][pathTemplate]);
        if (params) {
            var instance = new registry[services][pathTemplate];
            params.map(function (param) {
                this[param[0]] = param[1];
            }, instance);
            instance.attributes = registry[services][pathTemplate].attributes;
            if (exchanger) {
                exchanger.exchange(instance);
            }
            return instance;
        }
    }
    if (exchanger) {
        exchanger.exchange(undefined);
    }
    return undefined;
};

exports.serviceMatcher = serviceMatcher;

/**
 * Recursively evaluates any expressions in the metadata and reassigns to the value.
 */
var evalAttributes = function (attributes, request) {
    //Keep track of all promise creating attributes and only defer if any.
    var promises = [];
    if (Array.isArray(attributes)) {
        var result = [];
        for (var i = 0; i < attributes.length; i++) {
            evalAttributePromiseOrValue.call(this, attributes[i], promises, result, i, request);
        }
    } else {
        var result = {};
        for (var p in attributes) {
            evalAttributePromiseOrValue.call(this, attributes[p], promises, result, p, request);
        }
    }
    if (promises.length > 0) {
        return Promise.all(promises).then(function () {
            return result
        });
    } else {
        return result;
    }
};

/**
 * Evaluates the attribute and updates the parent object. If it's a promise the update will be deferred
 * and the promise will be added to the promises array.
 */
var evalAttributePromiseOrValue = function (attribute, promises, result, name, request) {
    var value = evalAttribute.call(this, attribute, request);
    if (Promise.isPromise(value)) {
        promises.push(value);
        value.then(function (value) {
            result[name] = value;
        });
    } else {
        result[name] = value;
    }
};

/**
 * Evaluates an attribute.
 */
var evalAttribute = function (attribute, request) {
    if (typeof attribute === 'function') {
        return attribute.call(this, request);
    } else if (typeof attribute === 'object' && !attribute.ELCompilerIgnore) {
        return evalAttributes.call(this, attribute, request);
    } else {
        return attribute;
    }
};

/**
 * Executes the shutdown hook registered in the service module javascript.
 */
var executeShutdownHook = function () {
    if (registry.shutdownHook) {
        registry.shutdownHook();
    }
};

exports.executeShutdownHook = executeShutdownHook;

var matchUrlTemplate = function (pathTemplate, path, cache) {
    if (cache && cache.regexp) {
        var regexp = cache.regexp;
    } else {
        pathTemplate = normalize(pathTemplate);
        // First replace wild cards with corresponding regexp ".*?"
        var regexpStr = pathTemplate.replace(/\*/g, ".*?");
        // Then replace all parameters "{{params}}" with regexp "(.*?)"
        regexpStr = regexpStr.replace(/\{\{.*?\}\}/g, "(.*?)");
        // Then replace all parameters "{param}"" with regexp "([^/]*?)"
        regexpStr = regexpStr.replace(/\{.*?\}/g, "([^/]*?)");
        // + signs have to be escaped in the regular expression
        regexpStr = "^" + regexpStr.replace(/\+/g, "\\+") + "$";
        var regexp = new RegExp(regexpStr);
        if (cache) {
            cache.regexp = regexp;
        }
    }
    path = normalize(path);
    // Execute regexp on path
    var result = regexp.exec(path);
    if (!result || result.length === 0) {
        return undefined;
    } else {
        result.shift();
        if (result.length > 0) {
            var paramRegexp = /\{([^{]*?)\}/g;  // TODO: Cache this regexp too?
            result = result.map(function (value) {
                var param = paramRegexp.exec(pathTemplate);
                return [param[1], value];
            });
        }
        log(result);
        return result;
    }
};

exports.matchUrlTemplate = matchUrlTemplate;

var serviceUrlMatcher = function (path, services, exchanger) {
    for (var pathTemplate in registry[services]) {
        var params = matchUrlTemplate(pathTemplate, path, registry[services][pathTemplate]);
        if (params) {
            if (exchanger) {
                exchanger.exchange(true);
            }
            return true;
        }
    }
    if (exchanger) {
        exchanger.exchange(false);
    }
    return false;
};

exports.serviceUrlMatcher = serviceUrlMatcher;

var matchRestOrPushService = function (path, exchanger) {
    var result = undefined;
    try {
        var instance = serviceMatcher(path, "restServices");
        if (instance) {
            result = {
                instance: instance,
                type: "rest"
            };
        } else if (instance = serviceMatcher(path, "pushServices")) {
            result = {
                instance: instance,
                type: "push"
            };
        }
    } catch (error) {
        log("Error while matching service: " + error);
    }
    exchanger.exchange(result);
};

exports.matchRestOrPushService = matchRestOrPushService;

var matchSocketService = function (path, exchanger) {
    var service = serviceMatcher(path, "socketServices");
    if (service) {
        Promise(evalAttributes.call(service, service.attributes)).then(function (attributes) {
            service.attributes = attributes;
            exchanger.exchange(service);
        }, function () {
            exchanger.exchange(undefined);
        });
    } else {
        exchanger.exchange(undefined);
    }
};

exports.matchSocketService = matchSocketService;

var findAndInvokeRestService = function (request, response) {
    var restService = serviceMatcher(request.url, "restServices");
    if (restService) {
        invokeRestService(restService, request, response);
    } else { // No match, return 404.
        response.setStatus(404);
        response.send(null, null);
    }
};

exports.findAndInvokeRestService = findAndInvokeRestService;

var invokeRestService = function (restService, request, response) {
    tryAndCatch(function () {
        var wrappedRequest = new WrappedRestRequest(request);
        var wrappedResponse = new WrappedRestResponse(response);
        return Promise(evalAttributes.call(restService, restService.attributes, wrappedRequest)).then(function (attributes) {
            restService.attributes = attributes;
            processCORSRules({
                                 origin: request.origin,
                                 method: request.accessControlRequestMethod,
                                 headers: Java.from(request.accessControlRequestHeaders)
                             }, request.method, wrappedResponse, restService.attributes.cors);
            if (restService.attributes.methods && restService.attributes.methods.indexOf(request.method.toUpperCase()) < 0) {
                wrappedResponse.status = 405;
                wrappedResponse.headers.Allow = arrayToString(restService.attributes.methods);
                return wrappedResponse.send();
            } else {
                return Promise(authorize(wrappedRequest, restService.attributes.authorization, restService)).then(function (authorized) {
                    if (!authorized) {
                        wrappedResponse.status = 403;
                        return wrappedResponse.send();
                    } else {
                        return restService.service(wrappedRequest, wrappedResponse);
                    }
                });
            }
        });
    }, function (error) {
        log("Failed to service REST request with URL " + request.url + ". " + error);
        if (error.dumpStack) {
            error.dumpStack();
        } else if (error.printStackTrace) {
            error.printStackTrace();
        }
        response.setStatus(500);
        response.send(null, null);
    });
};

exports.invokeRestService = invokeRestService;


/**
 * Parses an Origin header or an allowOrigin configuration parameter.
 * @param origin a string of the form "<protocol>://<host>:<port>" where protocol and port are optional.
 * @return An object with the properties "protocol" (optional), "host", "port" (optional).
 */
var parseOrigin = function (origin) {
    if (typeof origin !== 'string') {
        throw 'Origin must be a string';
    }
    var result = {};
    var split = origin.split('://', 2);
    if (split.length > 1) {
        result.protocol = split.shift() + ':';
    }
    split = split[0].split(':', 2);
    result.host = split.shift();
    if (split.length > 0) {
        result.port = split[0];
    }
    return result;
};

exports.parseOrigin = parseOrigin; // Used by the avatar module

/**
 * CORS policy configuration, used to enable cross-origin access to a restService or pushService.
 * @typedef CorsMetaData
 * @property {array} allowOrigin the list of origin this service allows. A single origin can be specified as just a string.
 * @property {array} exposeHeaders the list of response headers that may be exposed for a cross-origin request. A single header can be specified as just a string.
 * @property {boolean} allowCredentials whether or not user credentials are supported in a cross-origin request.
 * @property {array} allowMethods the list of methods that are allowed for a cross-origin request. A single method can be specified as just a string.
 * @property {array} allowHeaders the list of headers that are allowed for a cross-origin request. A single header can be specified as just a string.
 * @property {number} preflightMaxAge the number of seconds a preflight response is valid for.
 */

/**
 * Validates any CORS headers in the request and if it's a valid CORS request, adds corresponding headers to the response.
 * @param corsHeaders {object} the three request headers used for CORS processing: Origin, Access-Control-Request-Method, Access-Control-Request-Headers.
 * @param method the method of the request
 * @param response the response object
 * @param rules the CORS policy declared for the service
 */
var processCORSRules = function (corsHeaders, method, response, rules) {
    if (!corsHeaders.origin) { // 6.1.1
        return;
    }
    if (!rules || !rules.allowOrigin) {
        return;
    }
    var origin = parseOrigin(corsHeaders.origin);
    if (!origin.protocol) { // Invalid Origin header
        return;
    }
    if (!origin.port) {
        if (origin.protocol === 'https:') {
            origin.port = '443';
        } else if (origin.protocol === 'http:') {
            origin.port = '80';
        } // Any other protocol we just ignore the port since we don't know the default.
    }
    if (rules.allowOrigin != '*' && rules.allowOrigin.indexOf(origin.protocol + '//' + origin.host + ':' + origin.port) < 0) { // 6.1.2
        return;
    }
    if (method === 'Options') { // Preflight request
        if (!processPreflightRequest(corsHeaders, response, rules)) {
            return;
        }
    } else if (rules.exposeHeaders) {
        response.headers['Access-Control-Expose-Headers'] = rules.exposeHeaders;
    }
    if (rules.allowCredentials) { // 6.1.3
        response.headers['Access-Control-Allow-Credentials'] = 'true';
    }
    response.headers['Access-Control-Allow-Origin'] = corsHeaders.origin;
};

/**
 * @return false if it's an invalid preflight request, otherwise true.
 */
var processPreflightRequest = function (corsHeaders, response, rules) {
    if (!rules.allowMethods) {
        return false;
    }
    if (!corsHeaders.method || rules.allowMethods.indexOf(corsHeaders.method) < 0) { // 6.2.3, 6.2.5
        return false;
    }
    if (corsHeaders.headers) {
        if (!rules.allowHeaders) {
            return false;
        }
        if (corsHeaders.headers.length > 0) {
            for (var i = 0; i < corsHeaders.headers.length; i++) {
                var headerValues = corsHeaders.headers[i].split(',');
                for (var j = 0; j < headerValues.length; j++) {
                    if (rules.allowHeaders.indexOf(headerValues[j].trim().toLowerCase()) < 0) { // 6.2.6
                        return false;
                    }
                }
            }
            //No more return false below this point!
            response.headers['Access-Control-Allow-Headers'] = corsHeaders.headers; // 6.2.10 Return the list of requested headers
        }
    }
    response.headers['Access-Control-Allow-Methods'] = corsHeaders.method; // 6.2.9 Return the requested method
    if (rules.preflightMaxAge) {
        response.headers['Access-Control-Max-Age'] = rules.preflightMaxAge; // 6.2.8
    }
    return true;
};

var authorize = function (request, rules, service) {
    if (!rules) {
        return true; // No auth constraints means allow.
    }
    if (rules.allow) { //First match against the allow rules
        return rules.allow.map(function (allow) { //Convert the array of rules to functions, to pass into the promise chain.
            return function (previous) {
                if (previous) { //Previous rules has already allowed the request
                    return true;
                }
                return matchRule(allow, request, service);
            }
        }).reduce(function (p, f) {
                      return p.then(f);
                  }, Promise(false)) //Creates a promise chain of the functions in the array.
            .then(function (allowed) {
                      if (allowed) { //If any allow rules matched, go through the deny rules
                          if (rules.deny) {
                              return rules.deny.map(function (deny) {
                                  return function (previous) {
                                      if (!previous) { // A previous rule has already denied the request
                                          return false;
                                      }
                                      return !matchRule(deny, request, service);
                                  };
                              }).reduce(function (p, f) {
                                            return p.then(f);
                                        }, Promise(true)); //Initial value has to be true (allowed)
                          } else { // No deny rules -> allow.
                              return true;
                          }
                      } else { // No matching allow rules -> deny.
                          return false;
                      }
                  });
    }
    return false;
};

/**
 * A rules can consist of any combination of the following properties and for the rule to pass all of them have to match.
 * @typedef AuthRule
 * @property {array} methods A set of methods. The rule matches if the request method is equal to one of the methods in the list.
 * @property {string} user The name of the currently logged in user.
 * @property {string} group A group/role name. The rule matches if the currently logged in user is a member of this group.
 */

/**
 * Meta data for services to specify authorization constraints. If this metadata exists at least one of the allow rules
 * has to match in order for the request to be passed to the service. If no allow rules exist or at least one deny rule
 * matches a 403 will be returned.
 * @typedef AuthMetaData
 * @property {array} allow A set of {@link AuthRule AuthRule}s that will allow the request if matches.
 * @property {array} deny A set of {@link AuthRule AuthRule}s that will deny the request if matches.
 */

/**
 * Matches a request against a rule for a service.
 * A rule can have any of the following properties:
 *  "user" a username (compiled expression)
 *  "group" a group name
 *  "methods" an array of methods
 *  @returns true if the request matches the rule, false otherwise.
 */
var matchRule = function (rule, request, service) {
    if (rule.methods && rule.methods.indexOf(request.method.toUpperCase()) < 0) {
        return false;
    }
    if (rule.user) {
        if (request.user !== rule.user) {
            return false;
        }
    }
    if (rule.group) {
        return request.isUserInRole(rule.group).then(function (groupMatches) {
            return groupMatches;
        });
    }
    return true;
};

/**
 * @alias ServiceRequest
 * @constructor
 */
var WrappedServiceRequest = function (request) {
    /**
     * The authenticated user sending this request. If the request is not authenticated this property is null.
     */
    this.user = request.user;
    /**
     * The URL of the request.
     */
    this.url = request.url;

    /**
     * Asynchronously logs in a user. On a servlet container this implies creating a web session and associating the logged in
     * user with that session.
     *
     * @param {string} username the login name of the user
     * @param {string} password the login password
     * @param {function} callback (optional) callback that will be called when login has completed. The first
     * argument passed in to the callback is the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.login = function (username, password, callback) {
        var deferred = Promise.defer();
        request.login(username, password, function () {
            deferred.resolve();
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Asynchronously logs out the currently logged in user. If no user is logged in this causes the operation to fail.
     *
     * @param {function} callback (optional) callback that will be called when logout has been completed. The first
     * argument passed in to the callback is the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.logout = function (callback) {
        var deferred = Promise.defer();
        request.logout(function () {
            deferred.resolve();
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Asynchronously determines whether the currently logged in user has the give role. The result of this operation
     * is a boolean that has the value true if the used was logged in and has the given role, otherwise false.
     *
     * @param {string} role the role to test for
     * @param {function} callback (optional) callback that will be called when operation is completed. The first
     * argument passed in to the callback is the result and the second the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.isUserInRole = function (role, callback) {
        var deferred = Promise.defer();
        request.isUserInRole(role, function (event, args) {
            deferred.resolve(args[0]);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Asynchronously get the id of the current web session, if it exists.
     *
     * @param {function} callback (optional) callback that will be called when operation is completed. The first
     * argument passed in to the callback is the result and the second the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.getSessionId = function (callback) {
        var deferred = Promise.defer();
        request.getSessionId(function (event, args) {
            deferred.resolve(args[0]);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Asynchronously invalidate the current web session, if it exists.
     *
     * @param {function} callback (optional) callback that will be called when operation is completed. The first
     * argument passed in to the callback the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.invalidateSession = function (callback) {
        var deferred = Promise.defer();
        request.invalidateSession(function () {
            deferred.resolve();
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Asynchronously get a header value from the request.
     *
     * @param {string} header the name of the header to get.
     * @param {function} callback (optional) callback that will be called when operation is completed. The first
     * argument passed in to the callback is the result and the second the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.getHeader = function (header, callback) {
        var deferred = Promise.defer();
        request.getHeader(header, function (event, args) {
            deferred.resolve(args[0]);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Asynchronously get an array of header values from the request.
     *
     * @param {string} header the name of the header to get.
     * @param {function} callback (optional) callback that will be called when operation is completed. The first
     * argument passed in to the callback is the result and the second the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.getHeaders = function (header, callback) {
        var deferred = Promise.defer();
        request.getHeaders(header, function (event, args) {
            var result = Java.from(args);
            deferred.resolve(result);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };
};

/**
 * Rest Request object.
 * @alias RestRequest
 * @constructor
 * @extends ServiceRequest
 */
var WrappedRestRequest = function (request) {

    WrappedServiceRequest.apply(this, arguments);

    /**
     * A JavaScript object representing the deserialized content of the request.
     */
    this.data = request.data;
    /**
     * The content of the request in string form.
     */
    this.content = request.content;
    /**
     * The query string of the request.
     */
    this.query = request.query;
    /**
     * The request method.
     */
    this.method = request.method;
    /**
     * A Java object containing the Range header information.
     */
    this.range = request.range; // TODO: This should be moved to a JS object instead.
    /**
     * The value of the Content-Type header.
     */
    this.contentType = request.contentType;
};

exports.WrappedRestRequest = WrappedRestRequest;

/**
 * @alias ServiceResponse
 * @constructor
 */
var WrappedServiceResponse = function (response) {
    /**
     * A set of custom headers to include in the response.
     */
    this.headers = {};

    this.populateHeaders = function () {
        for (var header in this.headers) {
            if (Array.isArray(this.headers[header])) {
                for (var i = 0; i < this.headers[header].length; i++) {
                    response.addHeader(header, this.headers[header][i]);
                }
            } else {
                response.addHeader(header, this.headers[header]);
            }
        }
    };
};

/**
 * Rest Response object
 * @alias RestResponse
 * @constructor
 * @extends ServiceResponse
 */
var WrappedRestResponse = function (response) {

    //Inherit from WrappedServiceResponse
    WrappedServiceResponse.apply(this, arguments);

    /**
     * The method of the request for this response.
     */
    this.method = response.method;
    /**
     * The status code for this response.
     */
    this.status = undefined;
    /**
     * On object representing the content range of the response. The object can contain the properties
     * "count", "offset", and "total".
     */
    this.range;

    var doSend = function (response, result, callback) {
        response.send(result, function (event, args) {
            callback.apply(null, args);
        });
    };

    /**
     * Asynchronously sends the response. Unless [status]{@link WrappedRestResponse#status} has already been set by the user code
     * this method will set the status code depending on the value of the result parameter and the method.
     * If the result parameter is undefined, the status code will be set to 404.
     * If the result parameter is null the status code will be set to 204, except for GET requests where
     * the status code will be set to 404.
     * Otherwise the status code will be set to 200 and the result object will be serialized and set as content
     * of the response.
     *
     * If the
     *
     * @param {object} result (optional) result object to be included in the response.
     * @param {function} callback (optional) callback that will be called when the response has
     * been send. The first argument passed in to the callback is the error that occurred, if any.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.send = function (result, callback) {
        if (typeof result === "function") {
            callback = result;
        }
        this.populateHeaders();
        if (this.status) { //Set by application
            response.setStatus(this.status);
        } else if (result === undefined) { // Resource not found.
            response.setStatus(404);
        } else if (result === null) {
            if (this.method === "Get") { //TODO: Can get return 204?
                response.setStatus(404);
            } else {
                response.setStatus(204);
            }
        }
        if (result) {
            if (this.range) {
                response.setContentRange(this.range.count, this.range.offset, this.range.total);
            }
            if (typeof result !== "string") {
                result = JSON.stringify(result);
            }
            return wrapCallback(callback, nodeCall(doSend, response, result));
        } else {
            return wrapCallback(callback, nodeCall(doSend, response, null));
        }
    };
};

exports.WrappedRestResponse = WrappedRestResponse;

var normalize = function (path) {
    if (path.charAt(0) != "/") {
        path = "/" + path;
    }
    if (path.charAt(path.length - 1) != "/") {
        path = path + "/"
    }
    return path;
};

exports.normalize = normalize;

var stringify = function (object) {
    return JSON.stringify(object);
};

exports.stringify = stringify;

// -- Begin Push --
var findAndInvokePushService = function (url, pushServiceContext) {
    var instance = serviceMatcher(url, "pushServices");
    if (instance) {
        invokePushService(instance, url, pushServiceContext);
    } else {
        throw "No push service matches the url " + request.url;
    }
};

exports.findAndInvokePushService = findAndInvokePushService;

var invokePushService = function (pushService, url, pushServiceContext) {
    var request = pushServiceContext.request;
    tryAndCatch(function () {
        var wrappedResponse = new WrappedServiceResponse(pushServiceContext.response);
        var wrappedRequest = new WrappedServiceRequest(pushServiceContext.request);
        Promise(evalAttributes.call(pushService, pushService.attributes, wrappedRequest)).then(function (attribute) {
            pushService.attributes = attribute;
            processCORSRules({
                                 origin: request.origin,
                                 method: request.accessControlRequestMethod,
                                 headers: Java.from(request.accessControlRequestHeaders)
                             }, request.method, wrappedResponse, pushService.attributes.cors);
            wrappedResponse.populateHeaders();
            var wrappedContext = wrapPushContext(pushServiceContext, pushService);
            return Promise(authorize(wrappedRequest, pushService.attributes.authorization, pushService)).then(function (authorized) {
                if (!authorized) {
                    pushServiceContext.response.sendError(403);
                } else {
                    return pushService.onOpen(wrappedContext, wrappedRequest, wrappedResponse);
                }
            });
        });
    }, function (error) {
        log("Failed to call push service with URL " + url + ":" + error);
        pushServiceContext.response.sendError(500);
        error.dumpStack();
    });
};

exports.invokePushService = invokePushService;

/**
 * @name PushContext
 * @class
 * @classdesc A PushContext represents a connection from a browser to a {@link PushService}.
 *
 */
var wrapPushContext = function (pushServiceContext, instance) {
    var wrappedContext = {
        /**
         * Starts a timer for this context. When the timer expires the {@link PushService.onTimeout}
         * method will be called on the PushService this context belongs to.
         * @param timeout The number of milliseconds after which this timer will expire.
         * @memberof PushContext
         */
        setTimeout: function (timeout) {
            this.timeoutId = setTimeout(function () {
                instance.onTimeout(wrappedContext);
            }, timeout, this);
        },
        /**
         * Closes the connection represented by this context.
         * @memberof PushContext
         */
        close: function () {
            if (this.timeoutId) {
                clearTimeout(this.timeoutId);
            }
            instance.onClose(wrappedContext);
            pushServiceContext.close();
        },
        /**
         * Sends a message on this context's connection.
         * @param {object|string} message The message to send. If this is an object it will be serialized to JSON format.
         * @param {function} callback (optional) callback function to be called when the message has been sent.
         * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
         * @memberof PushContext
         */
        sendMessage: function (obj, callback) {
            if (typeof obj !== "string") {
                obj = JSON.stringify(obj);
            }
            var wrappedCallback;
            if (!callback) {
                // If no callback, return a promise
                var deferred = Promise.defer();
                wrappedCallback = function (name, args) { // ignore name
                    var err = args[0];
                    if (err) {
                        wrappedContext.close();
                        deferred.reject(new Error(err));
                    } else {
                        deferred.resolve();
                    }
                };
                pushServiceContext.sendMessage(obj, wrappedCallback);
                return deferred.promise;
            } else {
                wrappedCallback = function (name, args) { // ignore name
                    var err = args[0];
                    if (err) {
                        wrappedContext.close();
                    }
                    callback.apply(null, args);
                };
                pushServiceContext.sendMessage(obj, wrappedCallback);
            }
        }
    };

    return wrappedContext;
};

exports.wrapPushContext = wrapPushContext;

// -- Begin Socket --
var invokeSocketMethod = function (instance, name, peer, data) {
    peer = wrapPeer(peer);
    tryAndCatch(function () {
        if (data) {
            return instance[name](peer, data);
        } else {
            return instance[name](peer);
        }
    }, function (error) {
        log("Error while executing socket service method " + name + ": " + error);
        error.dumpStack();
    });
};

exports.invokeSocketMethod = invokeSocketMethod;

/**
 * @name Peer
 * @class
 * @classdesc A peer connected to a SocketService.
 *
 */
var wrapPeer = function (peer) {
    if (peer.getWrapper() == null) {
        peer.setWrapper({
                            /**
                             * Sends a message to this peer.
                             * @param {object|string} message The message to send. If this is an object it will be serialized to JSON format.
                             * @param {function} callback (optional) callback function to be called when the message has been sent.
                             * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
                             * @memberof Peer
                             */
                            send: function (obj, callback) {
                                if (typeof obj !== "string") {
                                    obj = JSON.stringify(obj);
                                }
                                var wrappedCallback;
                                if (!callback) {
                                    // If no callback, return a promise
                                    var deferred = Promise.defer();
                                    wrappedCallback = function (name, args) { // ignore name
                                        var err = args[0];
                                        if (err) {
                                            deferred.reject(new Error(err));
                                        } else {
                                            deferred.resolve();
                                        }
                                    };
                                    peer.send(obj, wrappedCallback);
                                    return deferred.promise;
                                } else {
                                    wrappedCallback = function (name, args) { // ignore name
                                        callback.apply(null, args);
                                    };
                                    peer.send(obj, wrappedCallback);
                                }
                            },
                            /**
                             * Gets the {@link SocketContext} that this Peer belongs to.
                             * @returns the SocketContext
                             * @memberof Peer
                             */
                            getContext: function () {
                                return wrapEndpointContext(peer.getContext());
                            }
                        });
    }
    return peer.getWrapper();
};

/**
 * @name SocketContext
 * @class
 * @classdesc A context representing a unique WebSocket URL
 *
 */
var wrapEndpointContext = function (context) {
    if (context.getWrapper() == null) {
        context.setWrapper({
                               /**
                                * Sends a message to all peers connected to the URL represented by this context.
                                * @param {object|string} message The message to send. If this is an object it will be serialized to JSON format.
                                * @param {function} callback (optional) callback function to be called when all messages have been sent.
                                * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
                                * @memberof SocketContext
                                */
                               sendAll: function (obj, callback) {
                                   if (typeof obj !== "string") {
                                       obj = JSON.stringify(obj);
                                   }
                                   var wrappedCallback;
                                   if (!callback) {
                                       // If no callback, return a promise
                                       var deferred = Promise.defer();
                                       wrappedCallback = function (name, args) { // ignore name
                                           var err = args[0];
                                           if (err) {
                                               deferred.reject(new Error(err));
                                           } else {
                                               deferred.resolve();
                                           }
                                       };
                                       context.sendAll(obj, wrappedCallback);
                                       return deferred.promise;
                                   } else {
                                       wrappedCallback = function (name, args) { // ignore name
                                           callback.apply(null, args);
                                       };
                                       context.sendAll(obj, wrappedCallback);
                                   }
                               },
                               /**
                                * Gets all Peers connected to this SocketContext.
                                * @returns an attay of all the {@link Peer} objects that were connected
                                * to URL of this SocketContext at the time this method was called.
                                * @memberof SocketContext
                                */
                               getPeers: function () {
                                   var peers = Java.from(context.getPeers());
                                   return peers.map(function (peer) {
                                       return wrapPeer(peer);
                                   });
                               }
                           });
    }
    return context.getWrapper();
};

exports.registry = registry;

/**
 * Converts an array of strings to a comma separated list.
 * @example ["A","B"] -> "A, B"
 */
var arrayToString = function (array) {
    var string = "";
    var comma = false;
    for (var i = 0; i < array.length; i++) {
        if (comma) {
            string = string + ', ';
        } else {
            comma = true;
        }
        string = string + array[i];
    }
    return string;
};


